@@ -2,7 +2,7 @@
  *                    Handles TCP port 1723 protocol.
  *                    C. Scott Ananian <cananian@alumni.princeton.edu>
  *
- * $Id: pptp_callmgr.c,v 1.5 2002/03/11 01:51:41 quozl Exp $
+ * $Id: pptp_callmgr.c,v 1.10 2003/06/11 08:17:29 quozl Exp $
  */
 #include <signal.h>
 #include <sys/time.h>
@@ -26,18 +26,20 @@
 #include "vector.h"
 #include "util.h"
 
+extern struct in_addr localbind; /* from pptp.c */
+
 int open_inetsock(struct in_addr inetaddr);
 int open_unixsock(struct in_addr inetaddr);
 void close_inetsock(int fd, struct in_addr inetaddr);
 void close_unixsock(int fd, struct in_addr inetaddr);
 
-sigjmp_buf env;
+sigjmp_buf callmgr_env;
 
-void sighandler(int sig) {
-    siglongjmp (env, 1);
+void callmgr_sighandler(int sig) {
+    siglongjmp (callmgr_env, 1);
 }
 
-void do_nothing(int sig) {
+void callmgr_do_nothing(int sig) {
     /* do nothing signal handler */
 }
 
@@ -51,20 +53,6 @@ struct local_conninfo {
   fd_set * call_set;
 };
 
-/* Connection callback */
-void conn_callback(PPTP_CONN *conn, enum conn_state state) {
-
-  switch(state) {
-  case CONN_OPEN_FAIL:
-  case CONN_CLOSE_DONE:
-    /* get outta here */
-    siglongjmp(env, 1);
-    break;
-  default:
-    log("Unhandled connection callback state [%d].", (int) state);
-    break;
-  }
-}
 /* Call callback */
 void call_callback(PPTP_CONN *conn, PPTP_CALL *call, enum call_state state) {
   struct local_callinfo *lci;
@@ -107,9 +95,9 @@ void call_callback(PPTP_CONN *conn, PPTP_CALL *call, enum call_state state) {
 
 /* NOTE ABOUT 'VOLATILE':                                              */
 /* several variables here get a volatile qualifier to silence warnings */
-/* from older (before 3.0) gccs. once the longjmp stuff is removed,    */
+/* from older (before 3.0) gccs. if the longjmp stuff is removed,      */
 /* the volatile qualifiers should be removed as well.                  */ 
-int main(int argc, char **argv, char **envp) {
+int callmgr_main(int argc, char **argv, char **envp) {
   struct in_addr inetaddr;
   int inet_sock, unix_sock;
   fd_set call_set;
@@ -144,15 +132,18 @@ int main(int argc, char **argv, char **envp) {
     exit(0);
   }
 
+  /* re-open stderr as /dev/null to release it */
+  file2fd("/dev/null", "wb", STDERR_FILENO);
+
   /* Step 1c: Clean up unix socket on TERM */
-  if (sigsetjmp(env, 1)!=0)
+  if (sigsetjmp(callmgr_env, 1)!=0)
     goto cleanup;
 
-  signal(SIGINT, sighandler);
-  signal(SIGTERM, sighandler);
+  signal(SIGINT, callmgr_sighandler);
+  signal(SIGTERM, callmgr_sighandler);
 
-  signal(SIGPIPE, do_nothing);
-  signal(SIGUSR1, do_nothing); /* signal state change; wake up accept */
+  signal(SIGPIPE, callmgr_do_nothing);
+  signal(SIGUSR1, callmgr_do_nothing); /* signal state change; wake up accept */
 
   /* Step 2: Open control connection and register callback */
   if ((conn = pptp_conn_open(inet_sock, 1, NULL/* callback */)) == NULL) {
@@ -172,7 +163,7 @@ int main(int argc, char **argv, char **envp) {
     pptp_conn_closure_put(conn, conninfo);
   }
 
-  if (sigsetjmp(env, 1)!=0) goto shutdown;
+  if (sigsetjmp(callmgr_env, 1)!=0) goto shutdown;
 
   /* Step 3: Get FD_SETs */
   do {
@@ -295,7 +286,7 @@ cleanup:
 }
 
 int open_inetsock(struct in_addr inetaddr) {
-  struct sockaddr_in dest;
+  struct sockaddr_in dest, src;
   int s;
 
   dest.sin_family = AF_INET;
@@ -306,12 +297,24 @@ int open_inetsock(struct in_addr inetaddr) {
     warn("socket: %s", strerror(errno));
     return s;
   }
+
+  if (localbind.s_addr != INADDR_NONE) {
+   bzero(&src, sizeof(src));
+   src.sin_family = AF_INET;
+   src.sin_addr   = localbind;
+   if (bind(s, (struct sockaddr *) &src, sizeof(src)) != 0) {
+     warn("bind: %s", strerror(errno));
+     close(s); return -1;
+   }
+  }
+
   if (connect(s, (struct sockaddr *) &dest, sizeof(dest)) < 0) {
     warn("connect: %s", strerror(errno));
     close(s); return -1;
   }
   return s;
 }
+
 int open_unixsock(struct in_addr inetaddr) {
   struct sockaddr_un where;
   struct stat st;
@@ -323,9 +326,7 @@ int open_unixsock(struct in_addr inetaddr) {
     return s;
   }
 
-  where.sun_family = AF_UNIX;
-  snprintf(where.sun_path, sizeof(where.sun_path), 
-	   PPTP_SOCKET_PREFIX "%s", inet_ntoa(inetaddr));
+  callmgr_name_unixsock(&where, inetaddr, localbind);
 
   if (stat(where.sun_path, &st) >= 0) {
     warn("Call manager for %s is already running.", inet_ntoa(inetaddr));
@@ -349,13 +350,27 @@ int open_unixsock(struct in_addr inetaddr) {
 
   return s;
 }
+
 void close_inetsock(int fd, struct in_addr inetaddr) {
   close(fd);
 }
+
 void close_unixsock(int fd, struct in_addr inetaddr) {
   struct sockaddr_un where;
+  
   close(fd);
-  snprintf(where.sun_path, sizeof(where.sun_path), 
-	   PPTP_SOCKET_PREFIX "%s", inet_ntoa(inetaddr));
+  callmgr_name_unixsock(&where, inetaddr, localbind);
   unlink(where.sun_path);
 }
+
+/* make a unix socket address */
+void callmgr_name_unixsock(struct sockaddr_un *where, 
+			   struct in_addr inetaddr,
+			   struct in_addr localbind) {
+  char localaddr[16], remoteaddr[16];
+  where->sun_family = AF_UNIX;
+  strncpy(localaddr,  inet_ntoa(localbind), 16);
+  strncpy(remoteaddr, inet_ntoa(inetaddr),  16);
+  snprintf(where->sun_path, sizeof(where->sun_path),
+	   PPTP_SOCKET_PREFIX "%s:%s", localaddr, remoteaddr);
+}
