@@ -1,7 +1,7 @@
 /* pptp_ctrl.c ... handle PPTP control connection.
  *                 C. Scott Ananian <cananian@alumni.princeton.edu>
  *
- * $Id: pptp_ctrl.c,v 1.10 2002/03/11 01:51:16 quozl Exp $
+ * $Id: pptp_ctrl.c,v 1.16 2003/05/23 19:06:10 reink Exp $
  */
 
 #include <errno.h>
@@ -23,6 +23,32 @@
 #include "util.h"
 #include "pptp_quirks.h"
 
+
+/* PPTP error codes: ----------------------------------------------*/
+
+/* (General Error Codes) */
+static const struct {
+  const char *name, *desc;
+} pptp_general_errors[] = {
+
+#define PPTP_GENERAL_ERROR_NONE                 0
+  { "(None)", "No general error" },
+#define PPTP_GENERAL_ERROR_NOT_CONNECTED        1
+  { "(Not-Connected)", "No control connection exists yet for this "
+                       "PAC-PNS pair" },
+#define PPTP_GENERAL_ERROR_BAD_FORMAT           2
+  { "(Bad-Format)", "Length is wrong or Magic Cookie value is incorrect" },
+#define PPTP_GENERAL_ERROR_BAD_VALUE            3
+  { "(Bad-Value)", "One of the field values was out of range or "
+                    "reserved field was non-zero" },
+#define PPTP_GENERAL_ERROR_NO_RESOURCE          4
+  { "(No-Resource)", "Insufficient resources to handle this command now" },
+#define PPTP_GENERAL_ERROR_BAD_CALLID           5
+  { "(Bad-Call ID)", "The Call ID is invalid in this context" },
+#define PPTP_GENERAL_ERROR_PAC_ERROR            6
+  { "(PAC-Error)", "A generic vendor-specific error occured in the PAC" }
+};
+
 /* BECAUSE OF SIGNAL LIMITATIONS, EACH PROCESS CAN ONLY MANAGE ONE
  * CONNECTION.  SO THIS 'PPTP_CONN' STRUCTURE IS A BIT MISLEADING.
  * WE'LL KEEP CONNECTION-SPECIFIC INFORMATION IN THERE ANYWAY (AS
@@ -105,6 +131,29 @@ struct PPTP_CALL {
   void * closure;
 };
 
+/* Outgoing Call Reply Result Codes */
+static const char *pptp_out_call_reply_result[] = {
+/* 0 */	"Unknown Result Code",
+/* 1 */	"Connected",
+/* 2 */	"General Error",
+/* 3 */	"No Carrier Detected",
+/* 4 */	"Busy Signal",
+/* 5 */	"No Dial Tone",
+/* 6 */	"Time Out",
+/* 7 */	"Not Accepted, Call is administratively prohibited" };
+
+#define MAX_OUT_CALL_REPLY_RESULT 7
+
+/* Call Disconnect Notify  Result Codes */
+static const char *pptp_call_disc_ntfy[] = {
+/* 0 */	"Unknown Result Code",
+/* 1 */	"Lost Carrier",
+/* 2 */	"General Error",
+/* 3 */	"Administrative Shutdown",
+/* 4 */	"(your) Request" };
+
+#define MAX_CALL_DISC_NTFY 4
+
 /* Local prototypes */
 static void pptp_reset_timer(void);
 static void pptp_handle_timer(int sig);
@@ -238,9 +287,10 @@ PPTP_CALL * pptp_call_open(PPTP_CONN * conn,
     /* fill in the phone number if it was specified */
     if (phonenr) {
         strncpy(packet.phone_num, phonenr, sizeof(packet.phone_num));
-        packet.phone_len = hton16 (strlen(phonenr));
+        packet.phone_len = strlen(phonenr);
         if( packet.phone_len > sizeof(packet.phone_num))
             packet.phone_len = sizeof(packet.phone_num);
+	packet.phone_len = hton16 (packet.phone_len);
     }
 
     if (pptp_send_ctrl_packet(conn, &packet, sizeof(packet))) {
@@ -679,17 +729,18 @@ void pptp_dispatch_ctrl_packet(PPTP_CONN * conn, void * buffer, size_t size) {
 	  unsigned int legal_error_value =
 	       sizeof(pptp_general_errors)/sizeof(pptp_general_errors[0]);
 	  int err = packet->error_code;
-          log("Error '%d' opening call. [callid %d]",
-	      packet->result_code, (int) callid);
-          log("Error code is '%d', Cause code is '%d'", err,
-	      packet->cause_code);
+          log("Our outgoing call request [callid %d] has not been accepted.", (int) callid);
+          log("Reply result code is %d '%s'. Error code is %d, Cause code is %d",
+	      packet->result_code,
+	      pptp_out_call_reply_result[packet->result_code <=  MAX_OUT_CALL_REPLY_RESULT ? packet->result_code : 0],
+	      err, packet->cause_code );
           if ((err > 0) && (err < legal_error_value)){
-            log("Error is '%s', Error message: '%s'",
-            pptp_general_errors[err].name,
-            pptp_general_errors[err].desc);
+	    if( packet->result_code != PPTP_RESULT_GENERAL_ERROR )
+		  log("Result code is something else then \"general error\", so the following error is probably bogus.");
+	    log("Error is '%s', Error message: '%s'",
+	    pptp_general_errors[err].name,
+	    pptp_general_errors[err].desc);
 	  }
-
-
 	  call->state.pns = PNS_IDLE;
 	  if (call->callback!=NULL) call->callback(conn, call, CALL_OPEN_FAIL);
 	  pptp_call_destroy(conn, call);
@@ -737,11 +788,26 @@ void pptp_dispatch_ctrl_packet(PPTP_CONN * conn, void * buffer, size_t size) {
 	(struct pptp_call_clear_ntfy *)buffer;
       if (vector_contains(conn->call, ntoh16(packet->call_id))) {
 	PPTP_CALL * call;
+	int err = packet->error_code;
+	unsigned int legal_error_value =
+	       sizeof(pptp_general_errors)/sizeof(pptp_general_errors[0]);
 	vector_search(conn->call, ntoh16(packet->call_id), &call);
+        log("Call disconnect notification received (call id %d)", (int) call->call_id);
+        log("Result code is %d: '%s'. Error code is %d, Cause code is %d",
+              packet->result_code,
+              pptp_call_disc_ntfy[packet->result_code <= MAX_CALL_DISC_NTFY ? packet->result_code : 0],
+              err, packet->cause_code );
+        if ((err > 0) && (err < legal_error_value)){
+            if( packet->result_code != PPTP_RESULT_GENERAL_ERROR )
+                log("Result code is something else then \"general error\", so the following error is probably bogus.");
+            log("Error is '%s', Error message: '%s'",
+            pptp_general_errors[err].name,
+            pptp_general_errors[err].desc);
+        }
 	pptp_call_destroy(conn, call);
-	log("Call closed (NTFY) (call id %d)", (int) call->call_id);
       }
       /* XXX we could log call stats here XXX */
+      /* XXX not all servers send this XXX */
       break;
     }
   case PPTP_SET_LINK_INFO:
@@ -753,7 +819,7 @@ void pptp_dispatch_ctrl_packet(PPTP_CONN * conn, void * buffer, size_t size) {
       if (ntoh32(packet->send_accm)==0 && ntoh32(packet->recv_accm)==0)
 	break; /* this is what we expect. */
       /* log it, otherwise. */
-      log("PPTP_SET_LINK_INFO recieved from peer_callid %u",
+      log("PPTP_SET_LINK_INFO received from peer_callid %u",
 	  (unsigned int) ntoh16(packet->call_id_peer));
       log("  send_accm is %08lX, recv_accm is %08lX",
 	  (unsigned long) ntoh32(packet->send_accm),
