@@ -2,7 +2,7 @@
  *                Handle the IP Protocol 47 portion of PPTP.
  *                C. Scott Ananian <cananian@alumni.princeton.edu>
  *
- * $Id: pptp_gre.c,v 1.9 2002/03/01 01:23:36 quozl Exp $
+ * $Id: pptp_gre.c,v 1.28 2003/04/15 12:49:16 quozl Exp $
  */
 
 #include <sys/types.h>
@@ -17,23 +17,30 @@
 #include <fcntl.h>
 #include "ppp_fcs.h"
 #include "pptp_msg.h"
+#include "pptp_gre.h"
 #include "util.h"
+#include "pqueue.h"
 
 #define PACKET_MAX 8196
 /* test for a 32 bit counter overflow */
 #define WRAPPED( curseq, lastseq) \
     ((((curseq) & 0xffffff00)==0) && (((lastseq) & 0xffffff00 )==0xffffff00))
+#undef REORDER_LOGGING
 
 static u_int32_t ack_sent, ack_recv;
 static u_int32_t seq_sent, seq_recv;
 static u_int16_t pptp_gre_call_id, pptp_gre_peer_call_id;
+gre_stats_t stats;
 
-/* decaps gets all the packets possible with ONE blocking read */
+typedef int (*callback_t)(int cl, void *pack, unsigned int len);
+
+/* decaps_hdlc gets all the packets possible with ONE blocking read */
 /* returns <0 if read() call fails */
-int decaps_hdlc(int fd, int (*cb)(int cl, void *pack, unsigned int len), int cl);
+int decaps_hdlc(int fd, callback_t callback, int cl);
 int encaps_hdlc(int fd, void *pack, unsigned int len);
-int decaps_gre (int fd, int (*cb)(int cl, void *pack, unsigned int len), int cl);
+int decaps_gre (int fd, callback_t callback, int cl);
 int encaps_gre (int fd, void *pack, unsigned int len);
+int dequeue_gre(callback_t callback, int cl);
 
 #if 1
 #include <stdio.h>
@@ -57,54 +64,126 @@ void print_packet(int fd, void *pack, unsigned int len) {
 }
 #endif
 
+uint64_t time_now_usecs() {
+  struct timeval tv;
+  gettimeofday(&tv, NULL);
+  return (tv.tv_sec * 1000000) + tv.tv_usec;
+}
+
+/* Open IP protocol socket */
+int pptp_gre_bind(struct in_addr inetaddr) {
+  struct sockaddr_in src_addr, loc_addr;
+  extern struct in_addr localbind;
+  
+  int s = socket(AF_INET, SOCK_RAW, PPTP_PROTO);
+  if (s < 0) { warn("socket: %s", strerror(errno)); return -1; }
+
+  if (localbind.s_addr != INADDR_NONE) {
+    bzero(&loc_addr, sizeof(loc_addr));
+    loc_addr.sin_family = AF_INET;
+    loc_addr.sin_addr   = localbind;
+    if (bind(s, (struct sockaddr *) &loc_addr, sizeof(loc_addr)) != 0) {
+      warn("bind: %s", strerror(errno)); close(s); return -1;
+    }
+  }
+
+  src_addr.sin_family = AF_INET;
+  src_addr.sin_addr   = inetaddr;
+  src_addr.sin_port   = 0;
+
+  if (connect(s, (struct sockaddr *) &src_addr, sizeof(src_addr)) < 0) {
+    warn("connect: %s", strerror(errno)); close(s); return -1;
+  }
+
+  return s;
+}
+
 void pptp_gre_copy(u_int16_t call_id, u_int16_t peer_call_id, 
-		   int pty_fd, struct in_addr inetaddr) {
-  struct sockaddr_in src_addr;
-  int s, n;
+		   int pty_fd, int gre_fd) {
+  int max_fd;
 
   pptp_gre_call_id = call_id;
   pptp_gre_peer_call_id = peer_call_id;
 
-  /* Open IP protocol socket */
-  s = socket(AF_INET, SOCK_RAW, PPTP_PROTO);
-  if (s<0) { warn("socket: %s", strerror(errno)); return; }
-  src_addr.sin_family = AF_INET;
-  src_addr.sin_addr   = inetaddr;
-  src_addr.sin_port   = 0;
-  if (connect(s, (struct sockaddr *) &src_addr, sizeof(src_addr))<0) {
-    warn("connect: %s", strerror(errno)); return;
-  }
   /* Pseudo-terminal already open. */
   
   ack_sent = ack_recv = seq_sent = seq_recv = 0;
 
-  n = (s>pty_fd)?(s+1):(pty_fd+1); /* weird select semantics */
+  /* weird select semantics */
+  max_fd = gre_fd;
+  if (pty_fd > max_fd) max_fd = pty_fd;
 
   /* Dispatch loop */
-  for (;;) { /* until error happens on s or pty_fd */
-    struct timeval tv = {0, 0}; /* non-blocking select */
+  for (;;) { /* until error happens on gre_fd or pty_fd */
+    struct timeval tv = {0, 0};
+    struct timeval *tvp;
     fd_set rfds;
     int retval;
+    pqueue_t *head;
+    int block_usecs = -1; /* wait forever */
 
     /* watch terminal and socket for input */
     FD_ZERO(&rfds);
-    FD_SET(s, &rfds);
-    FD_SET(pty_fd,&rfds);
+    FD_SET(gre_fd, &rfds);
+    FD_SET(pty_fd, &rfds);
 
-    /* if there is a pending ACK, do non-blocking select,
-       otherwise, block until data is available */
-    retval = select(n, &rfds, NULL, NULL, (ack_sent != seq_recv) ? &tv : NULL);
+    /*
+     * if there are multiple pending ACKs, then do non-blocking select;
+     * else if there is a single pending ACK then timeout after 0,5 seconds;
+     * else block until data is available.
+     */
 
-    if (retval == 0 && ack_sent != seq_recv) /* if outstanding ack */
-      encaps_gre(s, NULL, 0); /* send ack with no payload */
+    if (ack_sent != seq_recv) {
+      if (ack_sent + 1 == seq_recv)  /* u_int wrap-around safe */
+	block_usecs = 500000;
+      else
+	block_usecs = 0;
+    }
+    /* otherwise block_usecs == -1, which means wait forever */
 
-    if ((FD_ISSET(pty_fd, &rfds) && (decaps_hdlc(pty_fd, encaps_gre, s) < 0))
-     || (FD_ISSET(s,  &rfds) && (decaps_gre(s, encaps_hdlc, pty_fd) < 0)))
-        break;
+    /*
+     * If there is a packet in the queue, then don't wait longer than
+     * the time remaining until it expires.
+     */
+
+    head = pqueue_head();
+    if (head != NULL) {
+      int expiry_time = pqueue_expiry_time(head);
+      if (block_usecs == -1 || expiry_time < block_usecs)
+	block_usecs = expiry_time;
+    }
+
+    if (block_usecs == -1) {
+      tvp = NULL;
+    } else {
+      tvp = &tv;
+      tv.tv_usec = block_usecs;
+      tv.tv_sec  = tv.tv_usec / 1000000;
+      tv.tv_usec %= 1000000;
+    }
+
+    retval = select(max_fd+1, &rfds, NULL, NULL, tvp);
+
+    if (FD_ISSET(pty_fd, &rfds)) {
+	if (decaps_hdlc(pty_fd, encaps_gre,  gre_fd) < 0)
+	    break;
+    } else if (retval == 0 && ack_sent != seq_recv) {
+      /* if outstanding ack */
+      encaps_gre(gre_fd, NULL, 0);         
+      /* send ack with no payload */
+    }
+
+    if (FD_ISSET(gre_fd, &rfds)) {
+	if (decaps_gre (gre_fd, encaps_hdlc, pty_fd) < 0)
+	    break;
+    }
+    if (dequeue_gre (encaps_hdlc, pty_fd) < 0)
+      break;
   }
 
   /* Close up when done. */
-  close(s); close(pty_fd);
+  close(gre_fd);
+  close(pty_fd);
 }
 
 #define HDLC_FLAG         0x7E
@@ -121,14 +200,33 @@ int decaps_hdlc(int fd, int (*cb)(int cl, void *pack, unsigned int len), int cl)
 
   static unsigned int len = 0, escape = 0;
   static unsigned char copy[PACKET_MAX];
+  static int checkedsync = 0;
   
   /* start is start of packet.  end is end of buffer data */
   /*  this is the only blocking read we will allow */
 
   if ((end = read (fd, buffer, sizeof(buffer))) <= 0) {
-    log ("short read (%u): %s", end, strerror(errno));
+    warn("short read (%d): %s", end, strerror(errno));
+    if (errno == EIO) warn("pppd may have shutdown, see pppd log");
     return -1;
   }
+  /* FIXME: replace by warnings when the test is shown reliable */
+  if( !checkedsync) {
+      checkedsync = 1;
+      log( "PPP mode seems to be %s.\n",
+	      buffer[0] == HDLC_FLAG ? "Asynchronous" : "Synchronous" ); 
+  }
+  if ( syncppp ){
+      while ( start + 8 < end) {
+         len=ntoh16(*(short int*)(buffer+start+6))+4;
+         if ( start + len <= end) /* note: the buffer may contain an incomplete packet at the end
+                                   * this packet will be read again at the next read() */
+             if ((status = cb (cl, buffer + start, len)) < 0)
+                 return status; /* error-check */
+         start += len;
+      }
+      return 0;
+  }
   
   while (start < end) {
 
@@ -158,7 +256,7 @@ int decaps_hdlc(int fd, int (*cb)(int cl, void *pack, unsigned int len), int cl)
     }
     /* check, then remove the 16-bit FCS checksum field */
     if (pppfcs16 (PPPINITFCS16, copy, len) != PPPGOODFCS16)
-      log("Bad Frame Check Sequence during PPP to GRE decapsulation");
+      warn("Bad Frame Check Sequence during PPP to GRE decapsulation");
     len -= sizeof(u_int16_t);
 
     /* so now we have a packet of length 'len' in 'copy' */
@@ -180,6 +278,9 @@ int encaps_hdlc(int fd, void *pack, unsigned int len) {
   unsigned int pos=0, i;
   u_int16_t fcs;
 
+  if ( syncppp )
+      return write(fd, source, len);
+
   /* Compute the FCS */
   fcs = pppfcs16(PPPINITFCS16, source, len) ^ 0xFFFF;
 
@@ -205,14 +306,18 @@ int encaps_hdlc(int fd, void *pack, unsigned int len) {
   return write(fd, dest, pos);
 }
 
-int decaps_gre (int fd, int (*cb)(int cl, void *pack, unsigned int len), int cl) {
+int decaps_gre (int fd, callback_t callback, int cl) {
   unsigned char buffer[PACKET_MAX+64/*ip header*/];
   struct pptp_gre_header *header;
   int status, ip_len=0;
   static int first=1;
+  unsigned int headersize;
+  unsigned int payload_len;
+  u_int32_t seq;
 
   if ((status = read (fd, buffer, sizeof(buffer))) <= 0) {
-    log("short read (%u): %s", status, strerror(errno));
+    warn("short read (%d): %s", status, strerror(errno));
+    stats.rx_errors++;
     return -1;
   }
 
@@ -229,49 +334,123 @@ int decaps_gre (int fd, int (*cb)(int cl, void *pack, unsigned int len), int cl)
       (!PPTP_GRE_IS_K(ntoh8(header->flags))) || /* flag K should be set     */
       ((ntoh8(header->flags)&0xF)!=0)) { /* routing and recursion ctrl = 0  */
     /* if invalid, discard this packet */
-    log("Discarding GRE: %X %X %X %X %X %X", 
+    warn("Discarding GRE: %X %X %X %X %X %X", 
 	 ntoh8(header->ver)&0x7F, ntoh16(header->protocol), 
 	 PPTP_GRE_IS_C(ntoh8(header->flags)),
 	 PPTP_GRE_IS_R(ntoh8(header->flags)), 
 	 PPTP_GRE_IS_K(ntoh8(header->flags)),
 	 ntoh8(header->flags)&0xF);
+    stats.rx_invalid++;
     return 0;
   }
   if (PPTP_GRE_IS_A(ntoh8(header->ver))) { /* acknowledgement present */
     u_int32_t ack = (PPTP_GRE_IS_S(ntoh8(header->flags)))?
       header->ack:header->seq; /* ack in different place if S=0 */
+    ack = ntoh32( ack);
     if (ack > ack_recv) ack_recv = ack;
     /* also handle sequence number wrap-around  */
     if (WRAPPED(ack,ack_recv)) ack_recv=ack;
-  }
-  if (PPTP_GRE_IS_S(ntoh8(header->flags))) { /* payload present */
-    unsigned int headersize = sizeof(*header);
-    unsigned int payload_len= ntoh16(header->payload_len);
-    u_int32_t seq       = ntoh32(header->seq);
-    if (!PPTP_GRE_IS_A(ntoh8(header->ver))) headersize-=sizeof(header->ack);
-    /* check for incomplete packet (length smaller than expected) */
-    if (status-headersize<payload_len) return 0; 
-    /* check for out-of-order sequence number */
-    /* (handle sequence number wrap-around, and try to do it right) */
-    if ( first || (seq > seq_recv) || 
-	WRAPPED( seq, seq_recv)){
-      seq_recv = seq;
-      first=0;
-      return cb(cl, buffer+ip_len+headersize, payload_len);
-    } else {
-      log("discarding out-of-order  seq is %d seqrecv is %d", seq, seq_recv); 
-      return 0; /* discard out-of-order packets */
+
+    if (ack_recv == stats.pt.seq) {
+      int rtt = time_now_usecs() - stats.pt.time;
+      stats.rtt = (stats.rtt + rtt)/2;
     }
   }
-  return 0; /* ack, but no payload */
+
+  if (!PPTP_GRE_IS_S(ntoh8(header->flags))) /* no payload present */
+    return 0; /* ack, but no payload */
+
+  headersize  = sizeof(*header);
+  payload_len = ntoh16(header->payload_len);
+  seq         = ntoh32(header->seq);
+
+  if (!PPTP_GRE_IS_A(ntoh8(header->ver))) headersize-=sizeof(header->ack);
+
+  /* check for incomplete packet (length smaller than expected) */
+  if (status-headersize < payload_len) {
+    warn("discarding truncated packet (expected %d, got %d bytes)",
+	payload_len, status-headersize);
+    stats.rx_truncated++;
+    return 0; 
+  }
+
+  /* check for out-of-order sequence number */
+  /* (handle sequence number wrap-around, and try to do it right) */
+  if ( first || (seq == seq_recv+1) || 
+       WRAPPED(seq, seq_recv)){
+#if REORDER_LOGGING_VERBOSE
+    log("accepting packet %d", seq);
+#endif
+    stats.rx_accepted++;
+    first = 0;
+    seq_recv = seq;
+    return callback(cl, buffer+ip_len+headersize, payload_len);
+
+  } else if ( seq < seq_recv+1 || WRAPPED(seq_recv,seq) ) {
+    log("discarding duplicate or old packet %d (expecting %d)",
+	 seq, seq_recv+1);
+    stats.rx_underwin++;
+
+  } else if ( seq < seq_recv+MISSING_WINDOW ||
+	      WRAPPED(seq, seq_recv+MISSING_WINDOW) ) {
+#if 1 /* some logging of reordening is required, we might miss some "side effects" */
+    log("buffering out-of-order packet %d (expecting %d)", seq, seq_recv+1);
+#endif
+    pqueue_add(seq, buffer+ip_len+headersize, payload_len);
+    stats.rx_buffered++;
+
+  } else {
+    warn("discarding bogus packet %d (expecting %d)", seq, seq_recv+1);
+    stats.rx_overwin++;
+  }
+
+  return 0;
 }
+
+int dequeue_gre (callback_t callback, int cl) {
+  pqueue_t *head;
+  int status;
+
+  head = pqueue_head();
+
+  while ( head != NULL &&
+	  ( (head->seq == seq_recv+1)      || 
+	    (WRAPPED(head->seq, seq_recv)) ||
+	    (pqueue_expiry_time(head) <= 0) 
+	  )
+	) {
+
+    if (head->seq != seq_recv+1 && !WRAPPED(head->seq, seq_recv)) {
+      stats.rx_lost += head->seq - seq_recv - 1;
+#ifdef REORDER_LOGGING
+      log("timeout waiting for %d packets", head->seq - seq_recv - 1);
+#endif
+    }
+
+#ifdef REORDER_LOGGING
+    log("accepting %d from queue", head->seq);
+#endif
+
+    seq_recv = head->seq;
+    status = callback(cl, head->packet, head->packlen);
+    pqueue_del(head);
+    if (status < 0)
+      return status;
+
+    head = pqueue_head();
+  }
+
+  return 0;
+}
+
 int encaps_gre (int fd, void *pack, unsigned int len) {
   union {
     struct pptp_gre_header header;
     unsigned char buffer[PACKET_MAX+sizeof(struct pptp_gre_header)];
   } u;
-  static u_int32_t seq=0;
+  static u_int32_t seq = 1; /* first sequence number sent must be 1 */
   unsigned int header_len;
+  int rc;
 
   /* package this up in a GRE shell. */
   u.header.flags	= hton8 (PPTP_GRE_FLAG_K);
@@ -287,7 +466,15 @@ int encaps_gre (int fd, void *pack, unsigned int len) {
       u.header.payload_len = hton16(0);
       u.header.seq = hton32(seq_recv); /* ack is in odd place because S=0 */
       ack_sent = seq_recv;
-      return write(fd, &u.header, sizeof(u.header)-sizeof(u.header.seq));
+      rc = write(fd, &u.header, sizeof(u.header)-sizeof(u.header.seq));
+      if (rc < 0) {
+	stats.tx_failed++;
+      } else if (rc < sizeof(u.header)-sizeof(u.header.seq)) {
+	stats.tx_short++;
+      } else {
+	stats.tx_acks++;
+      }
+      return rc;
     } else return 0; /* we don't need to send ACK */
   } /* explicit brace to avoid ambiguous `else' warning */
   /* send packet with payload */
@@ -301,14 +488,27 @@ int encaps_gre (int fd, void *pack, unsigned int len) {
   } else { /* don't send ack */
     header_len = sizeof(u.header) - sizeof(u.header.ack);
   }
-  if (header_len+len>=sizeof(u.buffer)) return 0; /* drop this, it's too big */
+  if (header_len+len >= sizeof(u.buffer)) {
+    stats.tx_oversize++;
+    return 0; /* drop this, it's too big */
+  }
   /* copy payload into buffer */
   memcpy(u.buffer+header_len, pack, len);
   /* record and increment sequence numbers */
   seq_sent = seq; seq++;
   /* write this baby out to the net */
   /* print_packet(2, u.buffer, header_len+len); */
-  return write(fd, u.buffer, header_len+len);
+  rc = write(fd, u.buffer, header_len+len);
+  if (rc < 0) {
+    stats.tx_failed++;
+  } else if (rc < header_len+len) {
+    stats.tx_short++;
+  } else {
+    stats.tx_sent++;
+    stats.pt.seq  = seq_sent;
+    stats.pt.time = time_now_usecs();
+  }
+  return rc;
 }
 
 
