@@ -1,7 +1,7 @@
 /* util.c ....... error message utilities.
  *                C. Scott Ananian <cananian@alumni.princeton.edu>
  *
- * $Id: util.c,v 1.4 2002/03/11 01:50:07 quozl Exp $
+ * $Id: util.c,v 1.7 2002/12/09 05:50:37 quozl Exp $
  */
 
 #include <stdio.h>
@@ -15,6 +15,9 @@
 #define PROGRAM_NAME "pptp"
 #endif
 
+/* implementation of log_string, defined as extern in util.h */
+char *log_string = "anon";
+
 static void open_log(void) __attribute__ ((constructor));
 static void close_log(void) __attribute__ ((destructor));
 
@@ -30,8 +33,8 @@ va_list ap;						\
 char buf[256], string[256];				\
 va_start(ap, format);					\
 vsnprintf(buf, sizeof(buf), format, ap);		\
-snprintf(string, sizeof(string), "%s[%s:%s:%d]: %s",	\
-	 label, func, file, line, buf);			\
+snprintf(string, sizeof(string), "%s %s[%s:%s:%d]: %s",	\
+	 log_string, label, func, file, line, buf);	\
 va_end(ap)
 
 void _log(char *func, char *file, int line, char *format, ...) {
@@ -42,12 +45,26 @@ void _log(char *func, char *file, int line, char *format, ...) {
 void _warn(char *func, char *file, int line, char *format, ...) {
   MAKE_STRING("warn");
   fprintf(stderr, "%s\n", string);
-  syslog(LOG_NOTICE, "%s", string);
+  syslog(LOG_WARNING, "%s", string);
 }
 
 void _fatal(char *func, char *file, int line, char *format, ...) {
   MAKE_STRING("fatal");
   fprintf(stderr, "%s\n", string);
-  syslog(LOG_NOTICE, "%s", string);
+  syslog(LOG_CRIT, "%s", string);
   exit(1);
 }
+
+/* connect a file to a file descriptor */
+int file2fd(const char *path, const char *mode, int fd) {
+  int ok = 0;
+  FILE *file = NULL;
+
+  file = fopen(path, mode);
+  if (file != NULL && dup2(fileno(file), fd) != -1)
+    ok = 1;
+
+  if (file) fclose(file);
+
+  return ok;
+}
