@@ -1,12 +1,17 @@
 /* util.h ....... error message utilities.
  *                C. Scott Ananian <cananian@alumni.princeton.edu>
  *
- * $Id: util.h,v 1.1.1.1 2000/12/23 08:19:51 scott Exp $
+ * $Id: util.h,v 1.3 2002/12/09 05:50:37 quozl Exp $
  */
 
 #ifndef INC_UTIL_H
 #define INC_UTIL_H
 
+/* log_string is an identifier for this pptp process, passed from
+   command line using --log-string=X, and included with every log message.
+   Useful for people with multiple pptp sessions open at a time */
+extern char * log_string;
+
 void _log(char *func, char *file, int line, char *format, ...)
      __attribute__ ((format (printf, 4, 5)));
 void _warn(char *func, char *file, int line, char *format, ...)
@@ -21,4 +26,6 @@ void _fatal(char *func, char *file, int line, char *format, ...)
 #define fatal(format, args...) \
 	_fatal(__FUNCTION__,__FILE__,__LINE__, format , ## args)
 
+int file2fd(const char *path, const char *mode, int fd);
+
 #endif /* INC_UTIL_H */
